require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/finman/financialtransaction/credit_note_create'
require 'uu/finman/financialtransaction/credit_note_attributes'
require 'uu/finman/financialtransaction/credit_note_set_attributes'
require 'uu/finman/financialtransaction/control_side'
require 'uu/finman/financialtransaction/credit_note_type'

# Module Unicorn Universe
module UU

  # Module Finman
  module Finman

    # Module of finman Financial transaction
    module FinancialTransaction

      #Module of Credit Note
      module CreditNote
        extend self

        # Service path
        PATH = 'uu/finman/financialtransaction/FinancialTransaction'

        # Creates a new Credit Note. The command creates a new Credit Note for specified business case
        # with passed DTO parameters.
        #
        # <b>Required attributes:</b>
        #   control_business_case_uri, counterparty_business_case_uri, name, amount_without_vat, vat_rate or amount_with_vat, financial_transaction_type,  is_credit_note_outgoing,
        #   realization_probability, expected_pay_date, due_date, protocol_date, transaction_currency, (client_account_number and client_side_currency) or client_account_uri,
        #   (supplier_account_number and supplier_side_currency) or supplier_account_uri, vat_date
        # Exactly one of attributes <code>supplier_account_number</code> and <code>supplier_account_uri</code> has to be set.
        # Exactly one of attributes <code>client_account_number</code> and <code>client_account_uri</code> has to be set.
        # Exactly one of attributes <code>vat_rate</code> and <code>amount_with_vat</code> has to be set. Both attributes can be set simultaneously only if following statement is true : <code>amount_without_vat</code> * <code>vat_rate</code> = <code>amount_with_rate</code>.
        # Attribute <code>client_side_currency</code> and <code>supplier_side_currency</code> must be set if respective account is set by number.
        # Attribute <code>transaction_currency</code> must be set and must be one from <code>client_side_currency</code> and <code>supplier_side_currency</code>, with exception if both <code>client_side_currency</code> and <code>supplier_side_currency</code> are equal.
        #
        # @param [UU::OS::UESURI] control_business_case_uri UESURI of the business case of the control side to create Credit Note for.
        # @param [UU::Finman::FinancialTransaction::CreditNoteCreate] credit_note DTO containing attributes of the new Credit Note
        #
        # @return [UU::OS::UESURI] UESURI of the new Credit Note. In case of setting <code>dry_run</code> parameter to <code>true</code>, <code>UU::OS::UESURI::NIL_URI</code> is returned instead.
        #
        #
        # @example
        #      # Prepare Business Case URI, which we want to use to create Credit Note.
        #      control_business_case_uri = UU::OS::UESURI.new('ues:TERRITORY:BUSINESS_CASE');
        #
        #      # Create Credit note
        #      UU::Finman::FinancialTransaction::CreditNote.create(control_business_case_uri,
        #        :name => "Credit note",
        #        :supplier_description => "kreditní popis",
        #        :client_description => "debetní popis",
        #        :supplier_account_number => "123456",
        #        :client_account_number => "987654",
        #        :supplier_financial_category => UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/CR.TXWER"),
        #        :client_financial_category => UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/DR.TXOTR"),
        #        :realization_probability => 50,
        #        :isCreditNoteOutgoing => false,
        #        :financial_transaction_type => "NONCASH",
        #        :amount_without_vat => -10000,
        #        :vat_rate => 10,
        #        :invoice_number => "číslo faktury",
        #        :constant_symbol => "konstantní symbol",
        #        :variable_symbol => "variabilní symbol",
        #        :specific_symbol => "specifický symbol",
        #        :for_acceptance_date => "2014-11-02",
        #        :protocol_date => "2014-11-02",
        #        :Vat_date => "2014-11-02",
        #        :due_date => "2014-11-02",
        #        :expected_pay_date => "2014-11-02",
        #        :counterparty_business_case_uri =>  UU::OS::UESURI.new("ues:TERRITORY:CP_BC"),
        #        :client_account_number => "debetní číslo učtu",
        #        :is_planned => false,
        #        :is_ready_for_acceptance => true,
        #        :request_approval_by_manager => true,
        #        :wait_for_acceptance_by_accountant => true,
        #        :client_side_currency => "CZK",
        #        :supplier_side_currency => "EUR",
        #        :transaction_currency => "EUR"
        #       )
        def create(control_business_case_uri, credit_note)
          svc = UU::OS::REST::RemoteClient.new(CreditNote)
          credit_note_dto = UU::Finman::FinancialTransaction::CreditNoteCreate.new(credit_note)
          # adjust given payment dto to match expected format of financial transaction command
          transaction = transform_cnote_create_to_transaction_dto(credit_note_dto)
          financial_transaction_create = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', control_business_case_uri, financial_transaction_create)
            if res.nil?
              return res
            else
              return UU::OS::UESURI.new(res)
            end
          end
        end

        # Obtains Credit note attributes from service and returns them
        #
        # @param [String, UU::OS::UESURI] credit_note_uri UESURI of the Credit note
        #
        # @return [CreditNoteAttributes] Credit note attributes
        #
        #
        # @example
        #      # Prepare Credit note location which we want get attributes.
        #      credit_note_uri = UU::OS::UESURI.new('ues:TERRITORY:CREDIT_NOTE');
        #
        #      # Get Credit note
        #        credit_note_uesuri = UU::Finman::FinancialTransaction::CreditNote.get_attributes(credit_note_uri);
        def get_attributes(credit_note_uri)
          svc = UU::OS::REST::RemoteClient.new(CreditNote)
          UU::OS::QoS::QoSHandler.auto_retry do
            attrs_string = svc.get('getAttributes', credit_note_uri)
            attrs = JSON.parse(attrs_string, :symbolize_names => true)
            attrs = remove_invalid_keys(attrs)
            return UU::Finman::FinancialTransaction::CreditNoteAttributes.new(attrs)
          end
        end

        # Sets attributes of Credit note.
        #
        # @param [String,UU::OS::UESURI] credit_note_uri UESURI of modified Credit note
        # @param [CreditNoteSetAttributes] credit_note_set_attributes DTO containing modified attributes of the Credit note
        #
        # @return [UU::OS::UESURI] UESURI of modified Credit note
        #
        #
        # @example
        #      # Prepare Credit note location which we want set attributes.
        #      credit_note_uri = UU::OS::UESURI.new('ues:TERRITORY:CREDIT_NOTE');
        #
        #      # Set Credit note
        #        credit_note_uesuri = UU::Finman::FinancialTransaction::CreditNote.set_attributes(credit_note_uri,
        #        :name => "Credit note set");
        def set_attributes(credit_note_uri, credit_note_set_attributes)
          svc = UU::OS::REST::RemoteClient.new(CreditNote)
          credit_note_dto = UU::Finman::FinancialTransaction::CreditNoteSetAttributes.new(credit_note_set_attributes)
          # adjust given payment dto to match expected format of financial transaction command
          transaction = transform_cnote_modify_to_transaction_dto(credit_note_dto)
          financial_transaction_dto = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('setAttributes', credit_note_uri, financial_transaction_dto)
            return UU::OS::UESURI.new(res)
          end
        end

        # Deletes the specified credit note artifact from the system.
        # The credit note artifact cannot be deleted when the user executing this command is not cast in the Service Executives group.
        #
        # @param [String, UU::OS::UESURI] credit_note_uri UESURI of the credit_note which is about to be deleted.
        # 
        # @example
        #   # Delete credit note with specified URI.
        #   UU::Finman::FinancialTransaction::CreditNote.delete('ues:TERRITORY:P1234567890')
        def delete(credit_note_uri)
          svc = UU::OS::REST::RemoteClient.new(CreditNote)
          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('delete', credit_note_uri)
          end
        end        
        
        # Adjust dto of credit note to match requirements of financial transaction dto of server side.
        # @private
        def transform_cnote_create_to_transaction_dto(credit_note_dto)
          if credit_note_dto.is_a?(UU::OS::REST::DTO)
            attributes = credit_note_dto.to_hash
            # set class of transaction to determine Credit Note at the server side.
            attributes[:financialTransactionClass] = "CREDIT_NOTE"
            #transform attribute to another by attributes of financial transaction
            attributes[:isTransactionCredit] = attributes[:isCreditNoteOutgoing]
            attributes[:creditDescription] = attributes[:supplierDescription]
            attributes[:creditAccountUri] = attributes[:supplierAccountUri]
            attributes[:creditAccountNumber] = attributes[:supplierAccountNumber]
            attributes[:creditFinancialCategoryUri] = attributes[:supplierFinancialCategoryUri]
            attributes[:creditBusinessCaseFromOppositeTerritory] = attributes[:supplierBusinessCaseFromOppositeTerritory]
            attributes[:debitBusinessCaseFromOppositeTerritory] = attributes[:clientBusinessCaseFromOppositeTerritory]
            attributes[:debitDescription] = attributes[:clientDescription]
            attributes[:creditVatDate] = attributes[:vatDate]
            attributes[:debitAccountUri] = attributes[:clientAccountUri]
            attributes[:debitAccountNumber] = attributes[:clientAccountNumber]
            attributes[:debitFinancialCategoryUri] = attributes[:clientFinancialCategoryUri]
            attributes[:creditSideCurrency] = attributes[:supplierSideCurrency]
            attributes[:debitSideCurrency] = attributes[:clientSideCurrency]

            # remove unwanted attributes (in case that e.g. hashmap with attributes from other FinTx was modified and used)
            attributes.delete(:payDate)
            attributes.delete(:metaArtifactUri)
            attributes.delete(:forAcceptanceDate)
            attributes.delete(:isReadyForAcceptance)
            attributes.delete(:debitVatDate)
            attributes.delete(:isCreditNoteOutgoing)
            attributes.delete(:supplierDescription)
            attributes.delete(:clientDesscription)
            attributes.delete(:clientAccountUri)
            attributes.delete(:clientAccountNumber)
            attributes.delete(:clientBusinessCaseFromOppositeTerritory)
            attributes.delete(:supplierBusinessCaseFromOppositeTerritory)
            attributes.delete(:clientFinancialCategoryUri)
            attributes.delete(:vatDate)
            attributes.delete(:supplierAccountUri)
            attributes.delete(:supplierAccountNumber)
            attributes.delete(:supplierFinancialCategoryUri)
            attributes.delete(:isPlanned)
            attributes.delete(:supplierSideCurrency)
            attributes.delete(:clientSideCurrency)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_cnote_create_to_transaction_dto

        # Hash returned from service contains all attributes from financial transaction.
        # Remove invalid keys to contain only values for this specific class of transaction.
        # @private
        def remove_invalid_keys(hash)
          if hash.is_a?(Hash)
            hash[:supplierVatDate] = hash[:creditVatDate]
            hash[:supplierAccountNumber] = hash[:creditAccountNumber]
            hash[:supplierAccountUri] = hash[:creditAccountUri]
            hash[:supplierBusinessCaseFromOppositeTerritory] = hash[:creditBusinessCaseFromOppositeTerritory]
            hash[:supplierBusinessCaseUri] = hash[:creditBusinessCaseUri]
            hash[:supplierDescription] = hash[:creditDescription]
            hash[:supplierFinancialCategoryUri] = hash[:creditFinancialCategoryUri]
            hash[:clientAccountNumber] = hash[:debitAccountNumber]
            hash[:clientAccountUri] = hash[:debitAccountUri]
            hash[:clientBusinessCaseFromOppositeTerritory] = hash[:debitBusinessCaseFromOppositeTerritory]
            hash[:clientBusinessCaseUri] = hash[:debitBusinessCaseUri]
            hash[:clientDescription] = hash[:debitDescription]
            hash[:clientFinancialCategoryUri] = hash[:debitFinancialCategoryUri]
            hash[:clientVatDate] = hash[:debitVatDate]
            if (hash[:controlSide] == ControlSide::CREDIT)
              hash[:creditNoteType] = CreditNoteType::OUTGOING
            else
              hash[:creditNoteType] = CreditNoteType::INCOMING
            end

            hash[:clientSideCurrency] = hash[:debitSideCurrency]
            hash[:supplierSideCurrency] = hash[:creditSideCurrency]

            hash.delete(:acceptedDate)
            hash.delete(:complementaryTransferUri)
            hash.delete(:creditBusinessCaseFromOppositeTerritory)
            hash.delete(:creditBusinessCaseUri)
            hash.delete(:creditVatDate)
            hash.delete(:creditAccountNumber)
            hash.delete(:creditAccountUri)
            hash.delete(:creditDescription)
            hash.delete(:creditFinancialCategoryUri)
            hash.delete(:debitAccountNumber)
            hash.delete(:debitAccountUri)
            hash.delete(:debitBusinessCaseFromOppositeTerritory)
            hash.delete(:debitBusinessCaseUri)
            hash.delete(:debitDescription)
            hash.delete(:debitFinancialCategoryUri)
            hash.delete(:debitVatDate)
            hash.delete(:forAcceptanceDate)
            hash.delete(:isPlanned)
            hash.delete(:isReadyForAcceptance)
            hash.delete(:reinvoiceFromCategoryUri)
            hash.delete(:reinvoiceToCategoryUri)
            hash.delete(:reinvoiceFromItem)
            hash.delete(:reinvoiceOnItem)
            hash.delete(:reinvoicingType)
            hash.delete(:controlSide)
            hash.delete(:closeAfterCreation)
            hash.delete(:creditSideCurrency)
            hash.delete(:debitSideCurrency)
            return hash
          end
          fail "Given object must be Hash!"
        end
        private :remove_invalid_keys

        # Adjust dto of credit note to match requirements of financial transaction dto of server side.
        # @private
        def transform_cnote_modify_to_transaction_dto(credit_note_dto)
          if credit_note_dto.is_a?(UU::OS::REST::DTO)
            attributes = credit_note_dto.to_hash
            # set class of transaction to determine Credit Note at the server side.
            attributes[:financialTransactionClass] = "CREDIT_NOTE"

            #transform attribute to another by attributes of financial transaction
            attributes[:creditDescription] = attributes[:supplierDescription]
            attributes[:creditAccountUri] = attributes[:supplierAccountUri]
            attributes[:creditAccountNumber] = attributes[:supplierAccountNumber]
            attributes[:creditFinancialCategoryUri] = attributes[:supplierFinancialCategoryUri]
            attributes[:creditVatDate] = attributes[:supplierVatDate]
            attributes[:debitDescription] = attributes[:clientDescription]
            attributes[:debitAccountUri] = attributes[:clientAccountUri]
            attributes[:debitAccountNumber] = attributes[:clientAccountNumber]
            attributes[:debitFinancialCategoryUri] = attributes[:clientFinancialCategoryUri]
            attributes[:debitVatDate] = attributes[:clientVatDate]

            # remove unwanted attributes (in case that e.g. hashmap with attributes from other FinTx was modified and used)
            attributes.delete(:payDate)
            attributes.delete(:forAcceptanceDate)
            attributes.delete(:supplierDescription)
            attributes.delete(:clientDesscription)
            attributes.delete(:clientAccountUri)
            attributes.delete(:clientAccountNumber)
            attributes.delete(:clientVatDate)
            attributes.delete(:clientBusinessCaseFromOppositeTerritory)
            attributes.delete(:supplierBusinessCaseFromOppositeTerritory)
            attributes.delete(:clientFinancialCategoryUri)
            attributes.delete(:supplierVatDate)
            attributes.delete(:supplierAccountUri)
            attributes.delete(:supplierAccountNumber)
            attributes.delete(:supplierFinancialCategoryUri)
            attributes.delete(:isPlanned)
            attributes.delete(:requestApprovalByManager)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_cnote_modify_to_transaction_dto

      end
    end
  end
end
