require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/os/artifact/artifact_state_type'
require 'uu/finman/financialtransaction/financial_transaction_get_transaction_list'
require 'uu/finman/financialtransaction/financial_transaction_attributes'

# Module Unicorn Universe
module UU
  # Module Finman
  module Finman
    # Module of finman Financial transaction
    module FinancialTransaction
      extend self

      # Service path
      PATH = 'uu/finman/financialtransaction/FinancialTransaction'
      # Retrieves a result list of financial transactions for specified business case by UESURI.
      # The resulting attachment list is filtered according to the criteria provided by the criteria attribute.<br>
      #
      # Allowed values for attribute <code>businessCaseSide</code> in query are <code>CREDIT</code> for filtering by credit business case side and <code>DEBIT</code> for filtering by debit business case side. Use it for filtering transaction only by CREDIT / DEBIT side of Business Case(same as main entity). If parameter not used, returns transaction with CREDIT or DEBIT side.<br>
      # For allowed values for attribute <code>stateType</code> in query, look at UU::OS::Artifact::ArtifactStateType
      #
      # @param [String, UU::OS::UESURI] business_case_uri UESURI of the business case.
      # @param [UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionList] criteria DTO Financial_transaction_get_transaction_list containing paging and filtering criteria. Available parameters are:<br> {FinancialTransactionAttributes#creation_time creationTime}, {FinancialTransactionAttributes#credit_vat_date creditVatDate}, {FinancialTransactionAttributes#debit_vat_date debitVatDate}, {FinancialTransactionAttributes#meta_artifact_uri metaArtifactUri}, {FinancialTransactionAttributes#pay_date payDate}, {FinancialTransactionAttributes#state_name stateName}, {FinancialTransactionAttributes#state_type stateType}, {FinancialTransactionAttributes#cost_item_code costItemCode}, {FinancialTransactionAttributes#income_item_code incomeItemCode}, {FinancialTransactionAttributes#transaction_currency transactionCurrency}, businessCaseSide.
      # @option criteria [String] :query The UES query string. It has to conform the ues query syntax (see documentation for more info). Filtering and order of the result list can be specified there.
      #
      # @return [UU::OS::REST::ResultList<UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionList, UU::Finman::FinancialTransaction::FinancialTransactionAttributes>] List of financial transaction.
      #
      #
      # @example
      #      # Prepare Business Case URI, where we want to get Financial Transaction list.
      #      business_case_uri = UU::OS::UESURI.new('ues:TERRITORY:BUSINESS_CASE');
      #
      #      # Get transaction list on  specific Business Case
      #      UU::Finman::FinancialTransaction.get_transaction_list(business_case_uri,
      #         :query => "creationTime > '2015-01-01' AND metaArtifactUri = 'ues:TERRITORY:META_ARTIFACT_OF_FINANCIAL_TRANSACTION' AND stateType = ACTIVE");
      def get_transaction_list(bc_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(FinancialTransaction)
        dto = UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getTransactionList', bc_uri)
          return UU::OS::REST::ResultList.new(UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionList, UU::Finman::FinancialTransaction::FinancialTransactionAttributes, res)
        end
      end

      # Returns list of Financial Transactions related to the account by debit side that are matching the query criteria, specified by account UESURI.
      # The resulting attachment list is filtered according to the criteria provided by the criteria attribute.<br>
      #
      # Allowed values for attribute <code>transactionType</code> in query are <code>CASH</code> for show only NONCASH payments, <code>NONCASH</code> for show only NONCASH payments and <code>ALL</code> for both types. If parameter not used, returns transaction with type CASH or NONCASH .<br>
      #
      # @param [String, UU::OS::UESURI] account_uri UESURI of the account.
      # @param [UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionToBePaidList] criteria DTO Financial_transaction_get_transaction_to_be_paid_list containing paging and filtering criteria. Available parameters are:<br> {FinancialTransactionToBePaidAttributes#due_date dueDate, transactionType}.
      # @option criteria [String] :query The UES query string. It has to conform the ues query syntax (see documentation for more info). Filtering and order of the result list can be specified there.
      #
      # @return [UU::OS::REST::ResultList<UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionToBePaidList, UU::Finman::FinancialTransaction::FinancialTransactionToBePaidAttributes>] List of financial transactions.
      #
      #
      # @example
      #      # Prepare account URI, where we want to get Financial Transaction list.
      #      account_uri = UU::OS::UESURI.new('ues:TERRITORY:ACCOUNT_URI');
      #
      #      # Get transaction list on specific Account
      #      UU::Finman::FinancialTransaction.get_transaction_to_be_paid_list(account_uri,
      #         :query => "dueDate < '2016-01-01' AND dueDate > '2014-01-01'");
      #
      #      # Get transaction list on specific Account
      #      UU::Finman::FinancialTransaction.get_transaction_to_be_paid_list(account_uri,
      #         :query => "dueDate < '2016-01-01' AND dueDate > '2014-01-01' AND transactionType = 'CASH'");
      def get_transaction_to_be_paid_list(acc_uri, criteria = nil)
        svc = UU::OS::REST::RemoteClient.new(FinancialTransaction)
        dto = UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionToBePaidList.new(criteria)
        svc.add_parameter('pageIndex', dto.page_index)
        svc.add_parameter('pageSize', dto.page_size)
        svc.add_parameter('query', dto.query)
        UU::OS::QoS::QoSHandler.auto_retry do
          res = svc.get('getTransactionToBePaidList', acc_uri)
          return UU::OS::REST::ResultList.new(UU::Finman::FinancialTransaction::FinancialTransactionGetTransactionToBePaidList, UU::Finman::FinancialTransaction::FinancialTransactionToBePaidAttributes, res)
        end
      end

    end
  end
end
